from scipp import array, scalar, sqrt, vector
from scipp.spatial import rotations_from_rotvecs

from ess.spectroscopy.indirect import kf as secondary
from ess.spectroscopy.utils import norm


def vectors_close(a, b, tol=None):
    from scipp import dot, scalar, sqrt

    if tol is None:
        tol = scalar(1e-8, unit=a.unit)
    difference = a - b
    return sqrt(dot(difference, difference)) < tol


def all_vectors_close(a, b, tol=None):
    if len(a.shape) > 1 and len(b.shape) > 1:
        for d in a.dims:
            if d not in b.dims:
                return False
            if a.sizes[d] != b.sizes[d]:
                return False
        if a.shape != b.shape:
            return False
        return all_vectors_close(a.flatten(to='one'), b.flatten(to='one'), tol)
    if len(a.shape) and len(b.shape):
        if a.shape != b.shape:
            return False
        for x, y in zip(a, b, strict=True):
            if not vectors_close(x, y, tol):
                return False
    elif len(a.shape):
        for x in a:
            if not vectors_close(x, b, tol):
                return False
    elif len(b.shape):
        for y in b:
            if not vectors_close(a, y, tol):
                return False
    else:
        return vectors_close(a, b, tol)
    return True


def test_back_scattering_sample_analyzer_vector():
    sample_position = vector([0.0, 0.0, 0.0], unit='m')

    # The analyzer orientation defines the scattering plane coordinate system:
    analyzer_transform = rotations_from_rotvecs(vector([0.0, 0.0, 0.0], unit='degree'))

    x = analyzer_transform * vector([1.0, 0.0, 0.0], unit='1')
    y = analyzer_transform * vector([0.0, 1.0, 0.0], unit='1')
    z = analyzer_transform * vector([0.0, 0.0, 1.0], unit='1')
    # sample-analyzer-vector must be along the local z axis
    sample_analyzer_vec = scalar(1.0, unit='m') * z
    analyzer_position = sample_analyzer_vec + sample_position

    # analyzer orientation of 90 degrees ->
    # scattering angle of 180 degrees: back scattering
    analyzer_orientation = rotations_from_rotvecs(scalar(90.0, unit='degree') * y)

    detector_y = analyzer_orientation * analyzer_orientation * y
    assert vectors_close(detector_y, y)
    # The detector positions are defined in the coordinate system,
    # then rotated around the analyzer position
    tubes = array(values=[-0.1, 0.0, 0.1], unit='m', dims=['tube']) * x
    detector_positions = scalar(1.0, unit='m') * z + tubes
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )

    # This would be 'right' for fixed tau direction, but we utilize the mosaic and
    # insist on scattering from the center line of the analyzer
    # offset = sample_analyzer_vec - tubes / 2.0
    # (minus because the detector has been rotated 180 degrees)
    assert all_vectors_close(calculated, sample_analyzer_vec)

    # Offsetting along the analyzer scattering plane normal moves the analyzer
    # intersection point by half
    wires = array(values=[-0.1, 0.0, 0.1], unit='m', dims=['wire']) * y
    detector_positions = scalar(1.0, unit='m') * z + wires
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )
    assert all_vectors_close(calculated, sample_analyzer_vec + wires / 2.0)

    # Offsetting in both transverse directions at once
    detector_positions = scalar(1.0, unit='m') * z + wires + tubes
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )
    # as the x displacement increases, the y displacement should decrease
    frac = wires / (1.0 + sqrt(1.0 + (norm(tubes) / scalar(1.0, unit='m')) ** 2))
    assert all_vectors_close(calculated, sample_analyzer_vec + frac)


def test_sample_analyzer_vector():
    # TODO Randomize the sample position
    # TODO Randomize the analyzer transformation (a3 angle)
    # TODO Randomize the analyzer distance
    # TODO Randomize the detector positions

    from random import random

    sample_position = vector([0.0, 0.0, 0.0], unit='m')

    # The analyzer orientation defines the scattering plane coordinate system:
    analyzer_transform = rotations_from_rotvecs(vector([0.0, 0.0, 0.0], unit='degree'))

    x = analyzer_transform * vector([1.0, 0.0, 0.0], unit='1')
    y = analyzer_transform * vector([0.0, 1.0, 0.0], unit='1')
    z = analyzer_transform * vector([0.0, 0.0, 1.0], unit='1')
    # sample-analyzer-vector must be along the local z axis
    sample_analyzer_vec = scalar(0.5 + random(), unit='m') * z  # noqa: S311
    sample_analyzer_vec /= norm(sample_analyzer_vec).value
    analyzer_position = sample_analyzer_vec + sample_position

    # Avoid sin(theta) values near 0 or 1
    analyzer_orientation = rotations_from_rotvecs(scalar(45.0, unit='degree') * y)

    detector_y = analyzer_orientation * analyzer_orientation * y
    assert vectors_close(detector_y, y)
    # The detector positions are defined in the coordinate system,
    # then rotated around the analyzer position
    tubes = array(values=[-0.1, 0.0, 0.1], unit='m', dims=['tube']) * x
    detector_positions = scalar(1.0, unit='m') * z + tubes
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )
    offset = calculated - sample_analyzer_vec
    assert all_vectors_close(offset, 0 * tubes)

    # Offsetting along the analyzer scattering plane normal moves
    # the analyzer intersection point by half
    wires = array(values=[-0.1, 0.0, 0.1], unit='m', dims=['wire']) * y
    detector_positions = scalar(1.0, unit='m') * z + wires
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )
    assert all_vectors_close(calculated, sample_analyzer_vec + wires / 2.0)

    # Offsetting in both transverse directions at once
    detector_positions = scalar(1.0, unit='m') * z + wires + tubes
    detector_positions = (
        analyzer_position
        + analyzer_orientation * analyzer_orientation * detector_positions
    )
    calculated = secondary.sample_analyzer_vector(
        sample_position, analyzer_position, analyzer_orientation, detector_positions
    )
    # as the x displacement increases, the y displacement should decrease
    frac = wires / (1.0 + sqrt(1.0 + (norm(tubes) / scalar(1.0, unit='m')) ** 2))
    assert all_vectors_close(calculated, sample_analyzer_vec + frac)



from scipp import vector

from ..types import (
    EnergyTransfer,
    FinalWavenumber,
    FinalWavevector,
    IncidentWavenumber,
    IncidentWavevector,
    LabMomentumTransfer,
    LabMomentumTransferX,
    LabMomentumTransferY,
    LabMomentumTransferZ,
    SampleTableAngle,
    TableMomentumTransfer,
    TableMomentumTransferX,
    TableMomentumTransferY,
    TableMomentumTransferZ,
)
from .kf import providers as kf_providers
from .ki import providers as ki_providers

# Directions relative to the incident beam coordinate system
PERP, VERT, PARALLEL = (vector(v) for v in ([1, 0, 0], [0, 1, 0], [0, 0, 1]))


def lab_momentum_vector(
    ki: IncidentWavevector, kf: FinalWavevector
) -> LabMomentumTransfer:
    """Return the momentum transferred to the sample in the laboratory coordinate system

    The laboratory coordinate system is independent of sample angle

    Parameters
    ----------
    ki:
        incident wavevector of the neutron
    kf:
        final wavevector of the neutron

    Returns
    -------
    :
        The difference kf - ki
    """
    return kf - ki


def lab_momentum_x(q: LabMomentumTransfer) -> LabMomentumTransferX:
    """Return the X coordinate of the momentum transfer in the lab coordinate system"""
    from scipp import dot

    return dot(PERP, q)


def lab_momentum_y(q: LabMomentumTransfer) -> LabMomentumTransferY:
    """Return the Y coordinate of the momentum transfer in the lab coordinate system"""
    from scipp import dot

    return dot(VERT, q)


def lab_momentum_z(q: LabMomentumTransfer) -> LabMomentumTransferZ:
    """Return the Z coordinate of the momentum transfer in the lab coordinate system"""
    from scipp import dot

    return dot(PARALLEL, q)


def sample_table_momentum_vector(
    a3: SampleTableAngle, q: LabMomentumTransfer
) -> TableMomentumTransfer:
    """Rotate the momentum transfer vector into the sample-table coordinate system

    Notes
    -----
    When a3 is zero, the sample-table and lab coordinate systems are the same.
    That is, Z is along the incident beam, Y is opposite the gravitational force,
    and X completes the right-handed coordinate system. The sample-table angle, a3,
    has a rotation vector along Y, such that a positive 90-degree rotation places the
    sample-table Z along the lab X.

    Parameters
    ----------
    a3:
        The rotation angle of the sample table around the laboratory Y axis
    q:
        The momentum transfer in the laboratory coordinate system
    """
    from scipp.spatial import rotations_from_rotvecs

    # negative a3 since we rotate coordinates not axes here
    return rotations_from_rotvecs(-a3 * VERT) * q


def sample_table_momentum_x(q: TableMomentumTransfer) -> TableMomentumTransferX:
    """Return the X coordinate of the momentum transfer in the sample-table system"""
    from scipp import dot

    return dot(PERP, q)


def sample_table_momentum_y(q: TableMomentumTransfer) -> TableMomentumTransferY:
    """Return the Y coordinate of the momentum transfer in the sample-table system"""
    from scipp import dot

    return dot(VERT, q)


def sample_table_momentum_z(q: TableMomentumTransfer) -> TableMomentumTransferZ:
    """Return the Z coordinate of the momentum transfer in the sample-table system"""
    from scipp import dot

    return dot(PARALLEL, q)


def energy(ki: IncidentWavenumber, kf: FinalWavenumber) -> EnergyTransfer:
    """Calculate the energy transferred to the sample by a neutron"""
    from scipp.constants import hbar, neutron_mass

    return hbar * hbar * (ki * ki - kf * kf) / 2 / neutron_mass


providers = (
    *ki_providers,
    *kf_providers,
    lab_momentum_vector,
    lab_momentum_x,
    lab_momentum_y,
    lab_momentum_z,
    sample_table_momentum_vector,
    sample_table_momentum_x,
    sample_table_momentum_y,
    sample_table_momentum_z,
    energy,
)



from ess.spectroscopy.types import (
    AnalyzerDetectorVector,
    AnalyzerOrientation,
    AnalyzerPosition,
    DetectorFrameTime,
    DetectorGeometricA4,
    DetectorPosition,
    FinalEnergy,
    FinalWavenumber,
    FinalWavevector,
    ReciprocalLatticeSpacing,
    ReciprocalLatticeVectorAbsolute,
    SampleAnalyzerDirection,
    SampleAnalyzerVector,
    SampleDetectorFlightTime,
    SampleDetectorPathLength,
    SampleFrameTime,
    SamplePosition,
)


def sample_analyzer_vector(
    sample_position: SamplePosition,
    analyzer_position: AnalyzerPosition,
    analyzer_orientation: AnalyzerOrientation,
    detector_position: DetectorPosition,
) -> SampleAnalyzerVector:
    """Determine the sample to analyzer-reflection-point vector per detector element

    Note
    ----
    The shapes of the analyzer position and orientation should be self-consistent
    and will likely be 1:1. There is expected to be multiple detector element positions
    per analyzer which can be represented as an additional dimension compared to
    the analyzer shapes.

    Parameters
    ----------
    sample_position: scipp.DType.vector3
        The (probably singular) sample position, typically (0, 0, 0)
    analyzer_position: scipp.DType.vector3
        The nominal center of the central analyzer blade *surface*
    analyzer_orientation: scipp.DType.rotate3
        The orienting quaternion of the analyzer, used to identify the crystal y-axis
    detector_position: scipp.DType.vector3
        The position of the detector element

    Returns
    -------
    :
        The vector from the sample position to the interaction point on the analyzer
        for each detector element
    """
    from scipp import dot, vector

    from ess.spectroscopy.utils import norm

    # Scipp does not distinguish between coordinates and directions, so we need to do
    # some extra legwork to ensure we can apply the orientation transformation
    # _and_ obtain a dimensionless direction vector
    o = vector([0, 0, 0], unit=analyzer_orientation.unit)
    y = vector(
        [0, 1, 0], unit=analyzer_orientation.unit
    )  # and y perpendicular to the scattering plane
    yhat = analyzer_orientation * y - analyzer_orientation * o
    yhat /= norm(yhat)

    sample_analyzer_center_vector = analyzer_position - sample_position

    sample_detector_vector = detector_position - sample_position
    sd_out_of_plane = dot(sample_detector_vector, yhat)
    # the sample-detector vector is the sum of sample-analyzer,
    # analyzer-detector-center, the out-of-plane vector
    analyzer_detector_center_vector = (
        sample_detector_vector - sample_analyzer_center_vector - sd_out_of_plane * yhat
    )

    # TODO Consider requiring that dot(analyzer_position-sample_position, yhat) is zero?

    sample_analyzer_center_distance = norm(sample_analyzer_center_vector)

    analyzer_detector_center_distance = norm(analyzer_detector_center_vector)

    # similar-triangles give the out-of-plane analyzer reflection point distance
    sa_out_of_plane = (
        sample_analyzer_center_distance
        / (sample_analyzer_center_distance + analyzer_detector_center_distance)
        * sd_out_of_plane
    )

    return sample_analyzer_center_vector + sa_out_of_plane * yhat


def detector_geometric_a4(vec: SampleAnalyzerVector) -> DetectorGeometricA4:
    """Calculate the scattering angle from the incident beam to each detector element

    Parameters
    ----------
    vec : scipp.DType.vector3
        The per detector element analyzer interaction position, as determined
        by `sample_analyzer_vector`

    Returns
    -------
    :
        The per detector element scattering angle, a4, in degrees.
    """
    from scipp import atan2, dot, vector

    lab_x = vector(
        [1, 0, 0]
    )  # perpendicular to the incident beam, in the horizontal plane
    lab_z = vector([0, 0, 1])  # along the incident beam direction
    return atan2(y=dot(lab_x, vec), x=dot(lab_z, vec)).to(unit='deg')


def analyzer_detector_vector(
    sample_position: SamplePosition,
    sample_analyzer_vec: SampleAnalyzerVector,
    detector_position: DetectorPosition,
) -> AnalyzerDetectorVector:
    """Calculate the analyzer-detector vector"""
    return detector_position - (sample_position + sample_analyzer_vec)


def kf_hat(sample_analyzer_vec: SampleAnalyzerVector) -> SampleAnalyzerDirection:
    """Calculate the direction of the neutrons for each detector-element"""
    from ess.spectroscopy.utils import norm

    return sample_analyzer_vec / norm(sample_analyzer_vec)


def reciprocal_lattice_spacing(tau_vector: ReciprocalLatticeVectorAbsolute):
    """Calculate the distance between lattice planes in, e.g., the analyzer"""
    from ess.spectroscopy.utils import norm

    return norm(tau_vector)


def final_wavenumber(
    sample_analyzer_vec: SampleAnalyzerVector,
    analyzer_detector_vec: AnalyzerDetectorVector,
    tau: ReciprocalLatticeSpacing,
) -> FinalWavenumber:
    """Find the wave number of the neutrons reflected to each detector-element

    The wave number and wavelength are inversely proportional with
        wavelength = 2 * pi / wavenumber

    Parameters
    ----------
    sample_analyzer_vec : scipp.DType.vector3
        The vector from the sample to the analyzer interaction point for each
         detector element
    analyzer_detector_vec: scipp.DType.vector3
        The vector from the analyzer interaction point to its detector element,
        for each detector element
    tau: float-like
        The reciprocal lattice plane spacing of the analyzer crystal,
        likely in inverse angstrom

    Returns
    -------
    :
        The magnitude of the reflected neutron wave vector for each detector element
    """
    from scipp import sqrt

    from ess.spectroscopy.utils import norm

    # law of Cosines gives the scattering angle based on distances:
    l_sa = norm(sample_analyzer_vec)
    l_ad = norm(analyzer_detector_vec)
    l_diff = norm(sample_analyzer_vec + analyzer_detector_vec)
    # 2 theta is measured from the direction S-A, so the internal angle is
    # (pi - 2 theta) and the normal law of Cosines is modified accordingly to be
    # -cos(2 theta) instead of cos(pi - 2 theta)
    cos2theta = (l_diff * l_diff - l_sa * l_sa - l_ad * l_ad) / (2 * l_sa * l_ad)

    # law of Cosines gives the Bragg reflected wavevector magnitude
    return tau / sqrt(2 - 2 * cos2theta)


def final_energy(kf: FinalWavenumber) -> FinalEnergy:
    """Converts (final) wave number to (final) energy"""
    from scipp.constants import hbar, neutron_mass

    return ((hbar * hbar / 2 / neutron_mass) * kf * kf).to(unit='meV')


def final_wavevector(
    kf_direction: SampleAnalyzerDirection, kf_magnitude: FinalWavenumber
) -> FinalWavevector:
    """Constructs the final wave vector form its direction and magnitude"""
    return kf_direction * kf_magnitude


def secondary_flight_path_length(
    sample_analyzer_vec: SampleAnalyzerVector,
    analyzer_detector_vec: AnalyzerDetectorVector,
) -> SampleDetectorPathLength:
    """Returns the path-length-distance between the sample and each detector element"""
    from ess.spectroscopy.utils import norm

    return norm(sample_analyzer_vec) + norm(analyzer_detector_vec)


def secondary_flight_time(
    secondary_flight_distance: SampleDetectorPathLength, kf_magnitude: FinalWavenumber
) -> SampleDetectorFlightTime:
    """Calculates the most-likely time-of-flight between the sample and each pixel"""
    from scipp.constants import hbar, neutron_mass

    velocity = kf_magnitude * hbar / neutron_mass
    return secondary_flight_distance / velocity


def sample_frame_time(
    detector_time: DetectorFrameTime, secondary_time: SampleDetectorFlightTime
) -> SampleFrameTime:
    """Return the time each neutron likely interacted with the sample"""
    return detector_time - secondary_time


providers = (
    sample_analyzer_vector,
    analyzer_detector_vector,
    kf_hat,
    final_wavenumber,
    final_wavevector,
    secondary_flight_path_length,
    secondary_flight_time,
    sample_frame_time,
    final_energy,
    detector_geometric_a4,
)



from scipp import DataArray

from ..types import (
    FrameTimeMonitor,
    IncidentSlowness,
    MonitorNormalisation,
    PrimaryFocusDistance,
    PrimaryFocusTime,
    PrimarySpectrometerObject,
    SlownessMonitor,
    SourceFrequency,
    SourceMonitorFlightTime,
    SourceMonitorPathLength,
    WallTimeMonitor,
)


def incident_monitor_normalization(
    slowness: IncidentSlowness, monitor: SlownessMonitor
) -> MonitorNormalisation:
    """For each event, return the corresponding monitor intensity"""
    from scipp import lookup

    coords = list(monitor.coords)
    if len(coords) != 1:
        raise ValueError(f'Monitor expected to have exactly 1 coordinate, has {coords}')
    return lookup(monitor, dim=coords[0])[slowness]


def monitor_pivot_time(
    primary: PrimarySpectrometerObject, length: SourceMonitorPathLength
) -> SourceMonitorFlightTime:
    """Find the pivot time between source-pulse arrival times at the monitor position"""
    from choppera.nexus import primary_pivot_time_at

    return primary_pivot_time_at(primary, length)


def monitor_wall_time(
    monitor: FrameTimeMonitor,
    frequency: SourceFrequency,
    least: SourceMonitorFlightTime,
) -> WallTimeMonitor:
    """Convert the independent 'frame_time' coordinate of a histogram DataArray to
    the equivalent unwrapped 'wall_time'

    Parameters
    ----------
    monitor:
        A histogram beam monitor which has with recorded 'frame time' relative to
        the most-recent source pulse
    frequency:
        The source repetition frequency
    least:
        The minimum wall time between source pulse and arrival at the monitor position

    Returns
    -------
    :
        The same intensities with independent axis converted to the likely time since
        neutron-producing proton pulse
    """
    from choppera.nexus import unwrap

    coords = list(monitor.coords)
    if len(coords) != 1:
        raise ValueError(f'Monitor expected to have exactly 1 coordinate, has {coords}')
    frame = coords[0]
    wall = 'wall_time'
    names = {frame: wall}
    return DataArray(
        monitor.data.rename(names),
        coords={wall: unwrap(monitor.coords[frame], frequency, least).rename(names)},
    )


def monitor_slowness(
    monitor: WallTimeMonitor,
    length: SourceMonitorPathLength,
    distance: PrimaryFocusDistance,
    focus: PrimaryFocusTime,
) -> SlownessMonitor:
    """Convert the independent 'wall_time' coordinate of a histogram DataArray to
    the equivalent slowness

    Parameters
    ----------
    monitor:
        A histogram beam monitor which has been converted from recorded 'frame time' to
        time since producing proton pulse
    length:
        The path length from the source to the monitor
    distance:
        The distance from the source to the time-of-flight defining (chopper) position
    focus:
        The (mean) time from proton pulse to when all neutrons passed the tof
        defining point

    Returns
    -------
    :
        The same intensities with independent axis converted to the inverse velocity
        of the neutrons, which scales linearly with wall time
    """
    coords = list(monitor.coords)
    if len(coords) != 1:
        raise ValueError(f'Monitor expected to have exactly 1 coordinate, has {coords}')
    wall = coords[0]
    slow = 'slowness'
    names = {wall: slow}
    slowness = (
        ((monitor.coords[wall] - focus) / (length - distance))
        .rename(names)
        .to(unit='s/m')
    )
    return DataArray(monitor.data.rename(names), coords={slow: slowness})


providers = (
    incident_monitor_normalization,
    monitor_pivot_time,
    monitor_wall_time,
    monitor_slowness,
)



"""Utilities for the primary spectrometer of an
indirect geometry time-of-flight spectrometer
"""

from __future__ import annotations

from scippnexus import Group

from ess.spectroscopy.types import (
    FocusComponentName,
    FocusComponentNames,
    IncidentDirection,
    IncidentEnergy,
    IncidentSlowness,
    IncidentWavelength,
    IncidentWavenumber,
    IncidentWavevector,
    NeXusFileName,
    PrimaryFocusDistance,
    PrimaryFocusTime,
    PrimarySpectrometerObject,
    SampleFrameTime,
    SampleName,
    SamplePosition,
    SampleTime,
    SourceDelay,
    SourceDuration,
    SourceFrequency,
    SourceName,
    SourcePosition,
    SourceSampleFlightTime,
    SourceSamplePathLength,
    SourceVelocities,
)


def determine_name_with_type(
    instrument: Group, name: str | None, options: list, type_name: str
) -> str:
    """Investigate an open NeXus file group for objects with matching name or base type

    Parameter
    ---------
    instrument: scippnexus.Group
        The group to investigate, likely /entry/instrument
    name: str
        A preferred object name, will be returned if present in the scippnexus.Group
    options: list
        Should be scippnexus define NeXus object types, e.g., scippnexus.NXdetector
    type_name: str
        Candidate group members contain this name

    Returns
    -------
    :
        The matching group-member name

    Raises
    ------
    ValueError
        If no group member has the specified `name`,
        contains the specified `type_name` or is of the specified `option` type
    """
    if name is not None and name in instrument:
        return name
    found = {x for x in instrument if type_name in x.lower()}
    for option in options:
        found.update(set(instrument[option]))
    if len(found) != 1:
        raise ValueError(f"Could not determine {type_name} name: {found}")
    return next(iter(found))


def guess_source_name(file: NeXusFileName) -> SourceName:
    """Guess the name of the source in the NeXus instrument file"""
    from scippnexus import File, NXmoderator, NXsource

    with File(file) as data:
        instrument = data['entry/instrument']
        name = determine_name_with_type(
            instrument, None, [NXsource, NXmoderator], 'source'
        )
        return SourceName(name)


def guess_sample_name(file: NeXusFileName) -> SampleName:
    """Guess the name of the sample in the NeXus instrument file"""
    from scippnexus import File, NXsample

    with File(file) as data:
        instrument = data['entry/instrument']
        name = determine_name_with_type(instrument, None, [NXsample], 'sample')
        return SampleName(name)


def guess_focus_component_names(file: NeXusFileName) -> FocusComponentNames:
    """Guess the component names which define the focus of a Primary Spectrometer

    Note
    ----
    The order of components in the NeXus file must be consistent with the order of
    components along the beamline. This assumes that only NXdisk_chopper are used to
    define a focus distance, and that the first chopper or choppers along the beamline,
    within a fixed small distance, can define the focus distance. The component type,
    primacy, and allowed distance range could be user configurable inputs.

    Parameters
    ----------
    file: NeXusFileName
        The (HDF5) NeXus file name that contains an 'entry/instrument' group with
        one or more `scippnexus.NXdisk_chopper` groups inside

    Returns
    -------
    :
        The name or names of the time-focus-defining choppers, given the restrictions
        noted above
    """
    from scipp import scalar
    from scippnexus import File, NXdisk_chopper, compute_positions

    from ..utils import norm

    allowance = scalar(0.5, unit='m')

    with File(file) as data:
        instrument = data['entry/instrument']
        choppers = {
            k: compute_positions(v[...])['position']
            for k, v in instrument[NXdisk_chopper].items()
        }

    names = list(choppers.keys())
    focus_names = [FocusComponentName(names[0])]
    last = choppers[names[0]]['position']
    distance = 0 * allowance
    for name in names[1:]:
        x = choppers[name]['position']
        distance += norm(x - last)
        last = x
        if distance <= allowance:
            focus_names.append(FocusComponentName(name))
        else:
            break
    return FocusComponentNames(focus_names)


def source_position(file: NeXusFileName, source: SourceName) -> SourcePosition:
    """Extract the position of the named source from a NeXus file"""
    from scippnexus import File, compute_positions

    with File(file) as data:
        return compute_positions(data['entry/instrument'][source][...])['position']


def sample_position(file: NeXusFileName, sample: SampleName) -> SamplePosition:
    """Extract the position of the named sample from a NeXus file"""
    from scippnexus import File, compute_positions

    with File(file) as data:
        return compute_positions(data['entry/instrument'][sample][...])['position']


def focus_distance(
    file: NeXusFileName, origin: SourcePosition, names: FocusComponentNames
) -> PrimaryFocusDistance:
    """Find the average distance from the source position to the named components

    Warnings
    --------
    This distance is straight-line distance which may not precisely match the
    path-length that the neutrons take due to curved guides or other reflecting
    components. Care should be taken if the primary spectrometer includes any such
    components to ensure that the difference in flight path and straight-line path
    is not important

    Parameters
    ----------
    file: NeXusFileName
        The name of the HDF5 NeXus file containing the 'entry/instrument' group
        with the named components
    origin:
        The position of the source, likely obtained from the same NeXus file
    names: list
        The name(s) of the components whose average distance is calculated

    Returns
    -------
    :
        The average straight-line distance from the source position to the named
        component(s)
    """
    from scippnexus import File, compute_positions

    from ..utils import norm

    pos = 0 * origin
    with File(file) as data:
        for name in names:
            pos += compute_positions(data['entry/instrument'][name][...])['position']
    pos /= len(names)
    return norm(pos - origin)


def focus_time(
    primary: PrimarySpectrometerObject, distance: PrimaryFocusDistance
) -> PrimaryFocusTime:
    """Return the time relative to the pulse time that neutrons pass the focus"""
    from choppera.nexus import primary_focus_time

    return primary_focus_time(primary, distance)


def primary_path_length(
    file: NeXusFileName, source: SourcePosition, sample: SamplePosition
) -> SourceSamplePathLength:
    """Compute the primary spectrometer path length from source to sample positions

    Note:
        This *requires* that the instrument group *is sorted* along the beam path.
        HDF5 group entries are sorted alphabetically, so you should ensure that
        the NeXus file was constructed with this in mind.
    """
    from scipp import concat, dot, sqrt, sum
    from scippnexus import File, NXguide, compute_positions

    with File(file) as data:
        positions = [
            compute_positions(v[...])['position']
            for v in data['entry/instrument'][NXguide].values()
        ]

    positions = concat((source, *positions, sample), dim='path')
    diff = positions['path', 1:] - positions['path', :-1]
    return sum(sqrt(dot(diff, diff)))


def primary_spectrometer(
    file: NeXusFileName,
    source: SourceName,
    sample: SampleName,
    frequency: SourceFrequency,
    duration: SourceDuration,
    delay: SourceDelay,
    velocities: SourceVelocities,
) -> PrimarySpectrometerObject:
    """Construct and return a choppera.PrimarySpectrometer object

    Parameters
    ----------
    file:
        The HDF5 NeXus file with information about the primary spectrometer
        under 'entry/instrument'
    source:
        The name of the source component in the NeXus instrument group
    sample:
        The name of the sample component in the NeXus instrument group
    frequency:
        The frequency of the source, e.g., scipp.scalar(14.0, unit='Hz') for ESS
    duration:
        The source pulse duration, e.g., approximately scipp.scalar(3.0, unit='msec')
        for ESS
    delay:
        The velocity dependent time delay between source pulse time and neutrons
        exiting the moderator, should be 1-D (with dim='wavelength') and have values
        corresponding to the provided `velocities`
    velocities:
        The velocities of neutrons produced by the source; must be 1-D
        (with dim='wavelength') with at least two values. The velocities and delay
        times should be the same length, with a 1:1 correspondence. These values are
        then used to define the time-slowness phase space polygon produced by
        each source pulse.

    Returns
    -------
    :
        A choppera.PrimarySpectrometer object representing the NeXus file contents
    """
    from choppera.nexus import primary_spectrometer
    from scippnexus import File

    with File(file) as data:
        instrument = data['entry/instrument']
        if source not in instrument:
            raise KeyError(f"The source '{source}' is not in the instrument group")
        if sample not in instrument:
            raise KeyError(f"The sample '{sample}' is not in the instrument group")
        return primary_spectrometer(
            instrument, source, sample, frequency, duration, delay, velocities
        )


def primary_pivot_time(primary: PrimarySpectrometerObject) -> SourceSampleFlightTime:
    """Return a time, relative to the source pulse time, between neutron-arrival-periods

    Notes
    -----
    The determined pivot time is be before the earliest arrival time of neutrons
    produced in a source pulse that then are transmitted through the primary
    spectrometer, while also being after the latest arrival time from the preceding
    pulse.

    Parameters
    ----------
    primary:
        The choppera.PrimarySpectrometer object

    Returns
    -------
    :
        The determined pivot time
    """
    from choppera.nexus import primary_pivot_time as primary_time

    return primary_time(primary)


def unwrap_sample_time(
    times: SampleFrameTime, frequency: SourceFrequency, least: SourceSampleFlightTime
) -> SampleTime:
    """Use the pivot time to shift neutron event time offsets, recovering 'real'
    time after source pulse per event"""
    from choppera.nexus import unwrap as choppera_unwrap

    return choppera_unwrap(times, frequency, least)


def incident_slowness(
    length: SourceSamplePathLength,
    time: SampleTime,
    distance: PrimaryFocusDistance,
    focus: PrimaryFocusTime,
) -> IncidentSlowness:
    """Find the slowness of neutrons which pass through the primary spectrometer

    Parameters
    ----------
    length:
        The path length travelled by neutrons which go through the primary spectrometer
    time:
        The unwrapped 'wall' time for each event, relative to their producing source
        pulse
    distance:
        How far along the primary spectrometer the neutrons travelled before
        reaching the time-focus point
    focus:
        The 'wall' time relative to a source pulse time at which all neutrons must
        have passed the time-focus point

    Returns
    -------
    :
        The inverse of the velocity for each neutron, that is its 'slowness', which
        is proportional to wavelength
    """
    from ..utils import in_same_unit

    tof = time - in_same_unit(focus, to=time)
    slowness = tof / (length - distance)  # slowness _is_ inverse velocity
    return slowness


def incident_wavelength(slowness: IncidentSlowness) -> IncidentWavelength:
    """Calculate the incident wavelength from the incident slowness for each neutron"""
    from scipp.constants import Planck, neutron_mass

    return (slowness * Planck / neutron_mass).to(unit='angstrom')


def incident_wavenumber(slowness: IncidentSlowness) -> IncidentWavenumber:
    """Calculate the incident wave number from the incident slowness for each neutron"""
    from scipp.constants import hbar, neutron_mass

    return (neutron_mass / hbar / slowness).to(unit='1/angstrom')


def incident_direction() -> IncidentDirection:
    """Return the incident neutron direction in the laboratory frame"""
    from scipp import vector

    return vector([0, 0, 1.0])


def incident_wavevector(
    ki_magnitude: IncidentWavenumber, direction: IncidentDirection
) -> IncidentWavevector:
    """Find the incident wavevector from its magnitude and direction"""
    return ki_magnitude * direction


def incident_energy(ki: IncidentWavenumber) -> IncidentEnergy:
    """Convert the incident wavenumber to incident energy in meV"""
    from scipp.constants import hbar, neutron_mass

    return ((hbar * hbar / 2 / neutron_mass) * ki * ki).to(unit='meV')


providers = (
    sample_position,
    source_position,
    guess_sample_name,
    guess_source_name,
    guess_focus_component_names,
    focus_distance,
    focus_time,
    primary_path_length,
    primary_spectrometer,
    primary_pivot_time,
    unwrap_sample_time,
    incident_direction,
    incident_slowness,
    incident_wavelength,
    incident_wavenumber,
    incident_wavevector,
    incident_energy,
)



from typing import NewType

from choppera.primary import PrimarySpectrometer
from scipp import DataArray, Variable


def make_scipp_named_typer(scipp_type):
    def typer(named: str) -> type[scipp_type]:
        return NewType(named, scipp_type)

    return typer


variable_type = make_scipp_named_typer(Variable)
data_array_type = make_scipp_named_typer(DataArray)


NeXusFileName = NewType('NeXusFileName', str)

SourcePosition = variable_type('SourcePosition')
SamplePosition = variable_type('SamplePosition')
AnalyzerPosition = variable_type('AnalyzerPosition')
DetectorPosition = variable_type('DetectorPosition')
AnalyzerOrientation = variable_type('AnalyzerOrientation')
SampleAnalyzerVector = variable_type('SampleAnalyzerVector')
AnalyzerDetectorVector = variable_type('AnalyzerDetectorVector')
SampleAnalyzerDirection = variable_type('SampleAnalyzerDirection')
ReciprocalLatticeVectorAbsolute = variable_type('ReciprocalLatticeVectorAbsolute')
ReciprocalLatticeSpacing = variable_type('ReciprocalLatticeSpacing')
IncidentDirection = variable_type('IncidentDirection')
IncidentSlowness = variable_type('IncidentSlowness')
IncidentWavelength = variable_type('IncidentWavelength')
IncidentWavenumber = variable_type('IncidentWavenumber')
IncidentWavevector = variable_type('IncidentWavevector')
IncidentEnergy = variable_type('IncidentEnergy')
FinalDirection = variable_type('FinalDirection')
FinalSlowness = variable_type('FinalSlowness')
FinalWavelength = variable_type('FinalWavelength')
FinalWavenumber = variable_type('FinalWavenumber')
FinalWavevector = variable_type('FinalWavevector')
FinalEnergy = variable_type('FinalEnergy')
SourceSamplePathLength = variable_type('SourceSamplePathLength')
SourceSampleFlightTime = variable_type('SourceSampleFlightTime')
SampleDetectorPathLength = variable_type('SampleDetectorPathLength')
SampleDetectorFlightTime = variable_type('SampleDetectorFlightTime')
SampleFrameTime = variable_type('SampleFrameTime')
SampleTime = variable_type('SampleTime')
DetectorFrameTime = variable_type('DetectorFrameTime')
DetectorTime = variable_type('DetectorTime')
SourceTime = variable_type('SourceTime')

SourceName = NewType('SourceName', str)
SampleName = NewType('SampleName', str)
SourceFrequency = variable_type('SourceFrequency')
SourceDuration = variable_type('SourceDuration')
SourceDelay = variable_type('SourceDelay')
SourceVelocities = variable_type('SourceVelocities')

PrimarySpectrometerObject = NewType('PrimarySpectrometerObject', PrimarySpectrometer)

FocusComponentName = NewType('FocusComponentName', str)
FocusComponentNames = NewType('FocusComponentNames', list[FocusComponentName])
PrimaryFocusDistance = variable_type('PrimaryFocusDistance')
PrimaryFocusTime = variable_type('PrimaryFocusTime')

SourceMonitorPathLength = variable_type('SourceMonitorPathLength')
SourceMonitorFlightTime = variable_type('SourceMonitorFlightTime')
FrameTimeMonitor = data_array_type('FrameTimeMonitor')
WallTimeMonitor = data_array_type('WallTimeMonitor')
SlownessMonitor = data_array_type('SlownessMonitor')

MonitorNormalisation = variable_type('MonitorNormalisation')

LabMomentumTransfer = variable_type('LabMomentumTransfer')
LabMomentumTransferX = variable_type("LabMomentumTransferX")
LabMomentumTransferY = variable_type("LabMomentumTransferY")
LabMomentumTransferZ = variable_type("LabMomentumTransferZ")

SampleTableAngle = variable_type("SampleTableAngle")
TableMomentumTransfer = variable_type('TableMomentumTransfer')
TableMomentumTransferX = variable_type("TableMomentumTransferX")
TableMomentumTransferY = variable_type("TableMomentumTransferY")
TableMomentumTransferZ = variable_type("TableMomentumTransferZ")

SampleMomentumTransfer = variable_type('SampleMomentumTransfer')
SampleMomentumTransferX = variable_type("SampleMomentumTransferX")
SampleMomentumTransferY = variable_type("SampleMomentumTransferY")
SampleMomentumTransferZ = variable_type("SampleMomentumTransferZ")

EnergyTransfer = variable_type('EnergyTransfer')

# Debugging types, likely to be removed
DetectorGeometricA4 = variable_type("DetectorGeometricA4")



from scipp import DataArray, Variable


def norm(vector: Variable) -> Variable:
    from scipp import DType, dot, sqrt

    if vector.dtype != DType.vector3:
        raise ValueError("Vector operations require scipp.DType.vector3 elements")

    return sqrt(dot(vector, vector))


def in_same_unit(b: Variable, to: Variable | None = None) -> Variable:
    def unit(x):
        if x.bins is not None:
            return x.bins.unit
        return x.unit

    if to is None:
        raise ValueError("The to unit-full object must be specified")

    a_unit = unit(to)
    b_unit = unit(b)
    if a_unit is None and b_unit is None:
        return b
    if a_unit is None or b_unit is None:
        raise ValueError(f"Can not find the units to use for {b} from {to}")
    if a_unit != b_unit:
        b = b.to(unit=a_unit)
    return b


def is_in_coords(x: DataArray, name: str):
    return name in x.coords or (x.bins is not None and name in x.bins.coords)



from __future__ import annotations

from loguru import logger
from scipp import Variable

from ..types import NeXusFileName


def _load_all(group, obj_type):
    """Helper to find and load all subgroups of a specific scippnexus type"""
    return {name: obj[...] for name, obj in group[obj_type].items()}


def _load_named(group, obj_type, names):
    """Helper to find and load all subgroups of a specific scippnexus type with
    group name in an allowed set"""
    return {name: obj[...] for name, obj in group[obj_type].items() if name in names}


def ess_source_frequency():
    """Input for a sciline workflow, returns the ESS source frequency of 14 Hz"""
    from scipp import scalar

    return scalar(14.0, unit='Hz')


def ess_source_period():
    """Input for a sciline workflow, returns the ESS source period of 1/(14 Hz)"""
    return (1 / ess_source_frequency()).to(unit='ns')


def ess_source_delay():
    """Input for a sciline workflow, returns per-wavelength source delays of 0 s"""
    from scipp import array

    return array(values=[0, 0.0], dims=['wavelength'], unit='sec', dtype='float64')


def ess_source_duration():
    """Input for a sciline workflow, returns source pulse duration of 3 msec"""
    from scipp import scalar

    return scalar(3.0, unit='msec').to(unit='sec')


def ess_source_velocities():
    """Input for a sciline workflow, returns per-wavelength source velocity limits

    Notes
    -----
    The chosen limits are not based on any properties of the source, but rather
    entirely on the equivalent energy range, which is chosen to be
    ~53 micro electron volts to 530 milli electron volts. This energy range should
    be sufficient for all intended incident energies of the ESS spectrometer suite,
    but may not be sufficient to capture spurious energies that pass through
    the real instruments.

    Returns
    -------
    :
        A 1-D scipp Variable with values=[100, 10000] m/s
    """
    from scipp import array

    return array(values=[100, 1e4], dims=['wavelength'], unit='m/s')


def convert_simulated_time_to_frame_time(data):
    """Helper to make McStas simulated event data look more like real data

    McStas has the ability to track the time-of-flight from source to detector for
    every probabilistic neutron ray. This is very helpful, but unfortunately real
    instrument at ESS are not able to record the same information due to how the
    timing and data collection systems work.

    Real neutron events will record their event_time_zero most-recent-pulse reference
    time, and their event_time_offset detection time relative to that reference time.
    These two values added together give a real wall time; and information about the
    primary spectrometer is necessary to find any time-of-flight

    This function takes event data with per-event coordinate event_time_offset
    (actually McStas time-of-flight) and creates a new coordinate frame_time that is
    the time-of-flight modulo the source repetition period.

    Notes
    -----
    If the input data has realistic event_time_offset values, this function should
    produce frame_time data which is identical, and will therefore only increase
    the amount of data stored per event.

    Returns
    -------
    :
        A copy of the data with extra per-event coordinate frame_time
    """
    graph = {
        'frame_time': lambda event_time_offset: event_time_offset % ess_source_period()
    }
    return data.transform_coords(
        'frame_time', graph=graph, rename_dims=False, keep_intermediate=False
    )


def analyzer_per_detector(analyzers: list[str], triplets: list[str]) -> dict[str, str]:
    """Find the right analyzer name for each detector

    Notes
    -----
    Depends heavily on the names of components being preceded by an in-instrument index,
    and the analyzer and detector components being separated in index by 2.
    If either condition changes this function will need to be modified.

    Parameters
    ----------
    analyzers: list[str]
        The names of analyzer components, typically generated from the keys of a dict
    triplets: dict
        The names of triplet detector components, typically generated from the
        keys of a dict

    Returns
    -------
    :
        A dictionary with detector name keys and their associated analyzer name values
    """

    # TODO Update this function if the NeXus structure changes.
    def correct_index(d, a):
        detector_index = int(d.split('_', 1)[0])
        analyzer_index = detector_index - 2
        return a.startswith(str(analyzer_index))

    return {d: next(x for x in analyzers if correct_index(d, x)) for d in triplets}


def detector_per_pixel(triplets: dict) -> dict[int, str]:
    """Find the right detector name for every pixel index

    Parameters
    ----------
    triplets: dict[str, scipp.DataGroup]
        A mapping of detector component name to a loaded scippnexus.NXdetector group,
        with 'data' group member that has a 'detector_number' coordinate

    Returns
    -------
    :
        The mapping of detector_number to detector component name
    """
    return {
        i: name
        for name, det in triplets.items()
        for i in det['data'].coords['detector_number'].values.flatten()
    }


def combine_analyzers(analyzers: dict, triplets: dict):
    """Combine needed analyzer properties into a single array, duplicating information,
    to have per-pixel data

    BIFROST has 45 analyzers and 45 triplet detectors, each with some number of pixels,
    N. Calculations for the properties of neutrons which make it to each detector pixel
    need per-pixel data about the analyzer associated with that pixel. This function
    collects the required data and combines it into a single array.

    Analyzer information required to determine the secondary-spectrometer neutron
    properties are the center-of-mass position of the analyzer, the orientation of
    the analyzer, and the lattice spacing separating crystal planes in the analyzer.

    Notes
    -----
    Since there are N pixels per detector, the returned array is strictly N-times
    larger than necessary, but the optimization to use only minimal information is
    left for the future

    Parameters
    ----------
    analyzers: dict[str, scipp.DataGroup]
        Maps analyzer component name to loaded scippnexus.NXcrystal group
    triplets: dict[str, scipp.DataGroup]
        Maps detector component name to loaded scippnexus.NXdetector group

    Returns
    -------
    :
        A single array with 'event_id' pixel dimension and the
        per-pixel analyzer information
    """
    from scipp import Dataset, array, concat
    from scippnexus import compute_positions

    def analyzer_extract(obj):
        obj = compute_positions(obj, store_transform='transform')
        return Dataset(data={k: obj[k] for k in ('position', 'transform', 'd_spacing')})

    extracted = {k: analyzer_extract(v) for k, v in analyzers.items()}
    d2a = analyzer_per_detector(list(analyzers.keys()), list(triplets.keys()))
    p2d = detector_per_pixel(triplets)

    p2a = {k: extracted[d2a[v]] for k, v in p2d.items()}
    pixels = sorted(p2a)
    data = concat([p2a[p] for p in pixels], dim='event_id')
    data['event_id'] = array(values=pixels, dims=['event_id'], unit=None)
    return data


def combine_detectors(triplets: dict):
    """Combine needed detector properties into a single array

    BIFROST has 45 analyzers and 45 triplet detectors, each with some number of pixels,
    N. Calculations for the properties of neutrons which make it to each detector pixel
    need per-pixel data. This function collects the required data and combines it into
    a single array.

    Detector information required to determine the secondary-spectrometer neutron
    properties are the center-of-mass position of each pixel.

    Parameters
    ----------
    triplets: dict[str, scipp.DataGroup]
        Maps detector component name to loaded scippnexus.NXdetector group

    Returns
    -------
    :
        A single array with 'event_id' pixel dimension and the
        per-pixel center of mass position
    """
    from scipp import Dataset, concat, sort

    def extract(obj):
        pixels = obj['data'].coords['detector_number']
        midpoints = obj['data'].coords['position']
        return Dataset(data={'event_id': pixels, 'position': midpoints})

    data = concat([extract(v) for v in triplets.values()], dim='arm')
    data = Dataset({k: v.flatten(to='event_id') for k, v in data.items()})
    return sort(data, data['event_id'].data)


def find_sample_detector_flight_time(sample, analyzers, detector_positions):
    """Use sciline to find the sample to detector flight time per detector pixel"""
    import numpy as np
    from sciline import Pipeline

    from ..types import (
        AnalyzerOrientation,
        AnalyzerPosition,
        DetectorPosition,
        ReciprocalLatticeSpacing,
        SampleDetectorFlightTime,
        SamplePosition,
    )
    from .kf import providers as kf_providers

    params = {
        SamplePosition: sample['position'],
        AnalyzerPosition: analyzers['position'].data,
        DetectorPosition: detector_positions,  # detectors['position'].data,
        AnalyzerOrientation: analyzers['transform'].data,
        ReciprocalLatticeSpacing: 2 * np.pi / analyzers['d_spacing'].data,
    }
    return params, Pipeline(kf_providers, params=params).get(
        SampleDetectorFlightTime
    ).compute().to(unit='ms')


def get_triplet_events(triplets):
    """Extract and combine the events from loaded scippneutron.NXdetector groups

    Parameters
    ----------
    triplets:
        An iterable container of loaded NXdetector groups, each with a 'data' member
        which contains the pixel data -- possibly multiple detector-specific
        (but consistent) dimensions -- with a coordinate identifying the
        'detector_number'

    Returns
    -------
    :
       The events from each triplet concatenated and sorted by the 'detector_number'
    """
    from scipp import concat, sort

    events = concat([x['data'] for x in triplets], dim='arm').flatten(to='event_id')
    events = sort(events, events.coords['detector_number'])
    return events


def get_sample_events(triplet_events, sample_detector_flight_times):
    """Return the events with the frame_time coordinate offset to time at the sample"""
    events = triplet_events.copy()
    for coord in ('position', 'x_pixel_offset', 'y_pixel_offset'):
        del events.coords[coord]
    events.bins.coords['frame_time'] -= sample_detector_flight_times.to(unit='ns')
    events.bins.coords['frame_time'] %= ess_source_period()
    return events


def get_unwrapped_events(
    filename, source_name, sample_name, sample_events, focus_components
):
    """Shift frame_time at sample events to time-since-pulse events at sample"""
    from sciline import Pipeline

    from ..types import (
        FocusComponentNames,
        NeXusFileName,
        PrimarySpectrometerObject,
        SampleFrameTime,
        SampleName,
        SampleTime,
        SourceDelay,
        SourceDuration,
        SourceFrequency,
        SourceName,
        SourceVelocities,
    )
    from .ki import providers as ki_providers

    params = {
        NeXusFileName: filename,
        SampleName: sample_name,
        SourceName: source_name,
        SourceDelay: ess_source_delay(),
        SourceDuration: ess_source_duration(),
        SourceFrequency: ess_source_frequency(),
        SourceVelocities: ess_source_velocities(),
        SampleFrameTime: sample_events.data.bins.coords['frame_time'],
        FocusComponentNames: focus_components,
    }
    pipeline = Pipeline(ki_providers, params=params)
    primary = pipeline.get(PrimarySpectrometerObject).compute()

    events = sample_events.copy()
    events.bins.coords['frame_time'] = pipeline.get(SampleTime).compute()
    return params, events, primary


def get_normalization_monitor(monitors, monitor_component, collapse: bool = False):
    """Get the data of the named monitor component, converting frame_time to nanoseconds
    to match event_time_offset

    Parameters
    ----------
    monitors:
        A dictionary mapping monitor component name to loaded scippneutron.NXmonitor
        groups
    monitor_component:
        The name of the monitor component to access
    collapse: bool
        For some simulated experiments, a parameter was scanned which should not be
        treated as separate time points. When provided True, these points are integrated
        over the 'time' dimension.

    Returns
    -------
    :
        Monitor data with frame_time converted to nanoseconds to match the timescale
        used for events
    """
    normalization = monitors[monitor_component]['data']
    if collapse:
        # This is very specialized to how the simulated scans are done,
        # it needs to be generalized?
        normalization = normalization.sum(dim='time')
    # rescale the frame_time axis. Why does it need to be done this way?
    return normalization.transform_coords(
        ['frame_time'], graph={'frame_time': lambda t: t.to(unit='nanosecond')}
    )


def get_energy_axes(ki_params, kf_params):
    """Extract incident_energy, final_energy, and energy_transfer

    Parameters
    ----------
    ki_params:
        A dictionary of parameters needed by the incident-spectrometer sciline pipeline
    kf_params:
        A dictionary of parameters needed by the secondary-spectrometer sciline pipeline

    Returns
    -------
    ei:
        The incident energy
    en:
        The energy transfer
    ef:
        The final energy
    """
    from sciline import Pipeline

    from ..types import (
        EnergyTransfer,
        FinalEnergy,
        FinalWavenumber,
        IncidentEnergy,
        IncidentWavenumber,
    )
    from .conservation import providers

    params = {}
    params.update(ki_params)
    params.update(kf_params)
    pipeline = Pipeline(providers, params=params)
    pipeline[IncidentWavenumber] = pipeline.get(IncidentWavenumber).compute()
    pipeline[FinalWavenumber] = pipeline.get(FinalWavenumber).compute()
    ei = pipeline.get(IncidentEnergy).compute()
    en = pipeline.get(EnergyTransfer).compute()
    ef = pipeline.get(FinalEnergy).compute()
    return ei, en, ef


def add_momentum_axes(ki_params, kf_params, events, a3: Variable):
    """Extract momentum transfer in the lab and sample-table coordinate systems

    Parameters
    ----------
    ki_params:
        A dictionary of parameters needed by the incident-spectrometer sciline pipeline
    kf_params:
        A dictionary of parameters needed by the secondary-spectrometer sciline pipeline
    events:
        The event data to which the calculated momentum components are appended
    a3:
        The scalar value of the sample rotation angle describing the events

    Returns
    -------
    :
        The event data with the two horizontal plane components of the
        momentum transfer added in the laboratory coordinate system (independent of a3)
        and the sample-table coordinate system (rotated by a3 around y).
        These new coordinates are named 'lab_momentum_x', 'lab_momentum_z',
        'table_momentum_x' and 'table_momentum_z'
    """
    from sciline import Pipeline

    from ..types import (
        LabMomentumTransfer,
        LabMomentumTransferX,
        LabMomentumTransferZ,
        SampleTableAngle,
        TableMomentumTransfer,
        TableMomentumTransferX,
        TableMomentumTransferZ,
    )
    from .conservation import providers

    if a3.size != 1:
        raise ValueError(f'Expected a3 to have 1-entry, not {a3.size}')

    params = {}
    # First we must add the lab momentum vector, since it is not a3 dependent
    params.update(ki_params)
    params.update(kf_params)
    params[SampleTableAngle] = a3

    pipeline = Pipeline(providers, params=params)
    pipeline[LabMomentumTransfer] = pipeline.get(LabMomentumTransfer).compute()

    events.bins.coords['lab_momentum_x'] = pipeline.get(LabMomentumTransferX).compute()
    events.bins.coords['lab_momentum_z'] = pipeline.get(LabMomentumTransferZ).compute()

    pipeline[TableMomentumTransfer] = pipeline.get(TableMomentumTransfer).compute()
    events.bins.coords['table_momentum_x'] = (
        pipeline.get(TableMomentumTransferX).compute().transpose(events.dims)
    )
    events.bins.coords['table_momentum_z'] = (
        pipeline.get(TableMomentumTransferZ).compute().transpose(events.dims)
    )
    return events


def split(
    triplets,
    analyzers,
    monitors,
    logs,
    a3_name: str | None = None,
    a4_name: str | None = None,
):
    """Use the (a3, a4) logged value pairs to split triplet, analyzer, and monitor data
    into single-setting sets

    Parameters
    ----------
    triplets: DataArray
        The triplet positions _should_ change with a4 (if simulated, this is certainly
        not implemented correctly yet). The event data they contain depends on
        (a3, a4) [plus any other time-dependent parameter] so must be split.
    analyzers:
        The analyzer position _should_ change with a4
        (if simulated, this is certainly not implemented correctly yet)
    monitors: DataArray
        The histogram (simulated, or real current beam monitor) or event
        (real fission monitor, etc.?) data is time-dependent and used to normalize
        the detector data. It must therefore be split into (a3, a4) sets
    logs: DataGroup
        Entries built from NXlogs in the real or simulated instrument
    a3_name: str
        The name of the sample table angle log entry in `logs`, 'a3' for simulated data
    a4_name: str
        The name of the detector tank angle log entry in `logs`, 'a4' for simulated data

    Returns
    -------
    :
        A list[[triplet, analyzer, monitor]] of individual (a3, a4) setting(s)
    """
    from scipp import lookup

    from ..utils import is_in_coords

    if a3_name is None:
        a3_name = 'a3'
    if a4_name is None:
        a4_name = 'a4'

    if a3_name not in logs or a4_name not in logs:
        logger.warning('Missing a3 or a4, so split performed')
        return [[triplets, analyzers, monitors]]

    a3 = lookup(logs[a3_name]['value'], 'time')
    a4 = lookup(logs[a4_name]['value'], 'time')

    event_graph = {
        'a3': lambda event_time_zero: a3[event_time_zero],
        'a4': lambda event_time_zero: a4[event_time_zero],
    }
    histogram_graph = {'a3': lambda time: a3[time], 'a4': lambda time: a4[time]}

    def do_split(x, time_name):
        graph = event_graph if 'event_time_zero' == time_name else histogram_graph
        if is_in_coords(x, time_name):
            x = x.transform_coords(('a3', 'a4'), graph=graph)
            if x.bins is not None:
                x = x.group('a3', 'a4')
        return x

    vals = [
        do_split(x, t)
        for x, t in (
            (triplets, 'event_time_zero'),
            (analyzers, 'time'),
            (monitors, 'time'),
        )
    ]

    # FIXME this only works because v.sizes['a4'] is always 1 at the moment
    vals = [
        v.flatten(['a3', 'a4'], to='time') if 'a3' in v.dims and 'a4' in v.dims else v
        for v in vals
    ]

    n_time = [v.sizes['time'] for v in vals if 'time' in v.dims]
    if len(n_time):
        if not all(n == n_time[0] for n in n_time):
            raise ValueError("Not all values have the same 'time' dimension")
        n_time = n_time[0]
        vals = [
            [v['time', i] if 'time' in v.dims else v for v in vals]
            for i in range(n_time)
        ]
    else:
        vals = [vals]

    return vals


def load_everything(filename: NeXusFileName, named_components: dict[str, str]):
    """Load all needed information from the named NeXus HDF5 file

    Parameters
    ----------
    filename:
        The name of the file to load data from, must have both and 'instrument'
        and 'parameters' group under 'entry'
    named_components:
        The file-specific names of (at least) the source and sample group names
        under 'entry/instrument'

    Returns
    -------
    sample:
        The loaded sample component group
    triplets:
        All scippnexus.NXdetector groups under 'entry/instrument'
    analyzers:
        All scippnexus.NXcrystal groups under 'entry/instrument'
    choppers:
        All scippnexus.NXdisk_chopper groups under 'entry/instrument'
    monitors:
        All scippnexus.NXmonitor groups under 'entry/instrument'
    logs:
        The scippnexus.NXlog groups named 'a3' and 'a4' under 'entry/parameters'
    """
    import scippnexus as snx

    source_component = named_components['source']
    sample_component = named_components['sample']
    with snx.File(filename) as data:
        group = data['entry/instrument']
        if source_component not in group:
            raise ValueError(
                f'Missing {source_component=} for path-length calculations'
            )
        if sample_component not in group:
            raise ValueError(f'Missing {sample_component=} for origin identification')
        sample = snx.compute_positions(
            group[sample_component][...], store_transform='transform'
        )
        triplets = _load_all(group, snx.NXdetector)
        analyzers = _load_all(group, snx.NXcrystal)
        choppers = _load_all(group, snx.NXdisk_chopper)
        monitors = _load_all(group, snx.NXmonitor)

        # this is very BIFROST simulation specific -- can it be less so?
        # TODO use the _sample_ orientation itself to define a3 (and goniometer angles?)
        logs = _load_named(data['entry/parameters'], snx.NXlog, ('a3', 'a4'))

    return sample, triplets, analyzers, choppers, monitors, logs


def one_setting(
    sample, triplet_events, analyzers, norm_monitor, filename, names, warn_about_a3=True
):
    """Calculate the event properties for a single (a3, a4) setting"""
    detector_positions = triplet_events.coords['position']
    kf_params, sample_detector_flight_time = find_sample_detector_flight_time(
        sample, analyzers, detector_positions
    )
    sample_events = get_sample_events(triplet_events, sample_detector_flight_time)
    ki_params, unwrapped_events, primary = get_unwrapped_events(
        filename, names['source'], names['sample'], sample_events, names['focus']
    )
    ei, en, ef = get_energy_axes(ki_params, kf_params)

    energy_events = sample_events.copy()
    energy_events.bins.coords['energy_transfer'] = en.to(unit='meV')
    energy_events.bins.coords['incident_energy'] = ei
    energy_events.coords['final_energy'] = ef

    if 'a3' in triplet_events.coords:
        # this _should_ be one (a3, a4) setting,
        # with a single a3 value on triple_events (and norm_monitor)
        a3 = triplet_events.coords['a3']
    else:
        from scipp import scalar

        if warn_about_a3:
            logger.warning("No a3 present in setting, assuming 0 a3")
        a3 = scalar(0, unit='deg')

    energy_momentum_events = add_momentum_axes(ki_params, kf_params, energy_events, a3)

    return {
        'triplet_events': triplet_events,
        'sample_events': sample_events,
        'unwrapped_events': unwrapped_events,
        'norm_monitor': norm_monitor,
        'energy_events': energy_events,
        'sample_detector_flight_time': sample_detector_flight_time,
        'analyzers': analyzers,
        'energy_momentum_events': energy_momentum_events,
        # 'detectors': detectors,
        # 'monitors': monitors,
        # 'triplets': triplets,
    }


def load_precompute(
    filename: NeXusFileName,
    named_components: dict[str, str],
    is_simulated: bool = False,
):
    """Load data from a NeXus file and perform (a3, a4) independent calculations

    Parameters
    ----------
    filename:
        The file which contains the data to load
    named_components:
        The file-specific names of (at least) the source, sample and normalization
        monitor group names under 'entry/instrument'
    is_simulated:
        A flag to indicate if the file comes from a McStas simulation, such that
        the event_time_offset needs to be modified to look like real data.

    Returns
    -------
    sample:
        The sample group loaded from the NeXus file
    analyzers:
        A single array with all analyzer information needed to calculate
        secondary-spectrometer parameters
    triplet_events:
        A single array with the events and detector-pixel information needed
        to calculate parameters
    norm_monitor:
        The normalization monitor data with frame_time converted to nanoseconds
    logs:
        A dictionary of the 'a3' and 'a4' logs from the 'entry/parameter' group
        in the NeXus file
    """
    import scippnexus as snx

    sample, triplets, analyzers, choppers, monitors, logs = load_everything(
        filename, named_components
    )

    if is_simulated:
        for name in triplets:
            triplets[name]['data'] = convert_simulated_time_to_frame_time(
                triplets[name]['data']
            )

    for name in triplets:
        triplets[name] = snx.compute_positions(
            triplets[name], store_transform='transform'
        )

    analyzers = combine_analyzers(analyzers, triplets)
    # detectors = combine_detectors(triplets)
    triplet_events = get_triplet_events(triplets.values())

    norm_monitor = get_normalization_monitor(monitors, named_components['monitor'])
    return sample, analyzers, triplet_events, norm_monitor, logs


def component_names(
    source_component: str | None = None,
    sample_component: str | None = None,
    focus_components: list[str] | None = None,
    monitor_component: str | None = None,
    is_simulated: bool = False,
):
    """Return a dictionary mapping component type to component name

    Parameters
    ----------
    source_component: str
        The user-provided source component name, should exist at
        'entry/instrument/{source_component}' in the datafile
    sample_component: str
        The user-provided sample component name, should exist at
        'entry/instrument/{sample_component}' in the datafile
    focus_components: list[str]
        The user-provided set of component names defining the time-focus position,
        each should exist under 'entry/instrument' in the datafile
    monitor_component: str
        The user-provided normalization monitor component name, should exist at
        'entry/instrument/{monitor_component}'
    is_simulated: bool
        If true, user-provided names will be augmented with the McStas component
        names for the specific types.

    Returns
    -------
    :
        A dictionary mapping component type name to group name
    """
    from ..types import FocusComponentName

    names = {
        'source': source_component,
        'sample': sample_component,
        'focus': focus_components,
        'monitor': monitor_component,
    }
    if is_simulated:
        sim_components = {
            'source': '001_ESS_source',
            'sample': '114_sample_stack',
            'monitor': '110_frame_3',
            'focus': [
                FocusComponentName('005_PulseShapingChopper'),
                FocusComponentName('006_PulseShapingChopper2'),
            ],
        }
        for k, v in sim_components.items():
            if names[k] is None:
                names[k] = v
    return names


def bifrost(
    filename: NeXusFileName,
    source_component: str | None = None,
    sample_component: str | None = None,
    focus_components: list[str] | None = None,
    monitor_component: str | None = None,
    is_simulated: bool = False,
):
    """Load a BIFROST data file and convert to S(Q,E)
    in the sample-table coordinate system

    Parameters
    ----------
    filename:
        The name of the NeXus file to load
    source_component:
        The group name under 'entry/instrument' in the NeXus file containing
        source information
    sample_component:
        The group name under 'entry/instrument' in the NeXus file containing
        sample information
    focus_components:
        The group name or group names under 'entry/instrument' in the NeXus file
        which define the focus-time
    monitor_component:
        The group name under 'entry/instrument' in the NeXus file containing
        normalization monitor information
    is_simulated:
        Whether the NeXus file comes from a McStas simulation, in which case default
        component names are set if not provided and the data is modified to
        look like real data

    Returns
    -------
    A dictionary of data from the workflow, concatenated along a 'setting' dimension
    corresponding to separate (a3, a4) grouped data. The entries in the dictionary
    may not all be useful, and are subject to pruning as experience is gained with
    the workflow.
    """
    import scipp as sc
    from tqdm import tqdm

    named_components = component_names(
        source_component,
        sample_component,
        focus_components,
        monitor_component,
        is_simulated,
    )
    sample, analyzers, triplet_events, norm_monitor, logs = load_precompute(
        filename, named_components, is_simulated
    )
    settings = split(triplet_events, analyzers, norm_monitor, logs)
    data = [
        one_setting(
            sample,
            one_triplet_events,
            one_analyzers,
            one_monitor,
            filename,
            named_components,
        )
        for one_triplet_events, one_analyzers, one_monitor in tqdm(
            settings, desc='(a3, a4) settings'
        )
    ]
    return {k: sc.concat([d[k] for d in data], 'setting') for k in data[0]}


def bifrost_single(
    filename: NeXusFileName,
    source_component: str | None = None,
    sample_component: str | None = None,
    focus_components: list[str] | None = None,
    monitor_component: str | None = None,
    is_simulated: bool = False,
    extras: bool = False,
):
    """Load a BIFROST data file and convert to S(Q,E)
    in the laboratory coordinate system

    Parameters
    ----------
    filename:
        The name of the NeXus file to load
    source_component:
        The group name under 'entry/instrument' in the NeXus file containing
        source information
    sample_component:
        The group name under 'entry/instrument' in the NeXus file containing
        sample information
    focus_components:
        The group name or group names under 'entry/instrument' in the NeXus file
        which define the focus-time
    monitor_component:
        The group name under 'entry/instrument' in the NeXus file containing
        normalization monitor information
    is_simulated:
        Whether the NeXus file comes from a McStas simulation, in which case default
        component names are set if not provided and the data is modified to look like
        real data
    extras:
        If true, the loaded sample group and 'a3' and 'a4' logs will be returned in
        the dictionary

    Returns
    -------
    A dictionary of data from the workflow. The entries in the dictionary may not
    all be useful, and are subject to pruning as experience is gained with the workflow.
    """
    named_components = component_names(
        source_component,
        sample_component,
        focus_components,
        monitor_component,
        is_simulated,
    )
    sample, analyzers, triplet_events, norm_monitor, logs = load_precompute(
        filename, named_components, is_simulated
    )

    data = one_setting(
        sample,
        triplet_events,
        analyzers,
        norm_monitor,
        filename,
        named_components,
        warn_about_a3=False,
    )

    if extras:
        data['sample'] = sample
        data['logs'] = logs

    return data
